#!/usr/bin/env python

from random import random, randrange
from os.path import splitext
import EXIF
import PIL.Image
import itertools
import math
import sys

corners = [
    (0, 0, 0), (255, 0, 0), (0, 255, 0), (255, 255, 0),
    (0, 0, 255), (255, 0, 255), (0, 255, 255), (255, 255, 255)
    ]

def within(v, lower, upper):
    return min(max(v, lower), upper)

def clip(v):
    return within(v, 0, 255)

def distance(rgb1, rgb2):
    r1, g1, b1 = rgb1
    r2, g2, b2 = rgb2
    return math.sqrt(pow(r1 - r2, 2) + pow(g1 - g2, 2) + pow(b1 - b2, 2))

def nearest(rgb, values):
    l = [(distance(rgb, value), value) for value in values]
    l.sort()
    return l[0][1]

def restore(srgb):
    return tuple(map(lambda x: int(x * 255.0), srgb))

def scale(rgb):
    return tuple(map(lambda x: x / 255.0, rgb))

def square(srgb):
    return tuple(map(lambda x: pow(x, 2), srgb))

def invert(srgb):
    return tuple(map(lambda x: 1.0 - x, srgb))

def combination(rgb):
    rgb = square(scale(rgb))
    rgbi = invert(rgb)
    pairs = zip(rgbi, rgb)
    d = []
    for corner in corners:
        rs, gs, bs = scale(corner)
        d.append((pairs[0][int(rs)] * pairs[1][int(gs)] * pairs[2][int(bs)], corner))
    return balance(d)

def complements(rgb):
    r, g, b = rgb
    return rgb, restore(invert(scale(rgb)))

def balance(d):
    d = dict([(value, f) for f, value in d])
    for primary, secondary in map(complements, [(255, 0, 0), (0, 255, 0), (0, 0, 255)]):
        common = min(d[primary], d[secondary])
        d[primary] -= common
        d[secondary] -= common
        d[(0, 0, 0)] += common
        d[(255, 255, 255)] += common
    return [(f, value) for value, f in d.items()]

def compensate(d, chosen):
    dd = dict([(value, f) for f, value in d])
    for f, value in d:
        if value not in chosen:
            _value, complement = complements(value)
            if complement in chosen:
                f = max(0, f - dd[complement])
                dd[value] = f
    return [(f, value) for value, f in dd.items() if value in chosen]

def combine(d):
    out = [0, 0, 0]
    for v, rgb in d:
        out[0] += v * rgb[0]
        out[1] += v * rgb[1]
        out[2] += v * rgb[2]
    return out

def pattern(rgb, chosen=None):
    l = combination(rgb)
    if chosen:
        l = compensate(l, chosen)
    l.sort(reverse=True)
    return l

def get_value(rgb, chosen=None):
    l = pattern(rgb, chosen)
    limit = sum([f for f, c in l])
    choose = random() * limit
    threshold = 0
    for f, c in l:
        threshold += f
        if choose < threshold:
            return c
    return c

def sign(x):
    return x >= 0 and 1 or -1

def saturate_rgb(rgb, exp):
    return tuple([saturate_value(x, exp) for x in rgb])

def saturate_value(x, exp):
    return int(127.5 + sign(x - 127.5) * 127.5 * pow(abs(x - 127.5) / 127.5, exp))

def amplify_rgb(rgb, exp):
    return tuple([amplify_value(x, exp) for x in rgb])

def amplify_value(x, exp):
    return int(pow(x / 255.0, exp) * 255.0)

def get_colours(im, y):
    width, height = im.size
    c = {}
    for x in range(0, width):
        rgb = im.getpixel((x, y))

        # Sum the colour probabilities.

        for f, value in combination(rgb):
            if not c.has_key(value):
                c[value] = f
            else:
                c[value] += f

    c = [(n/width, value) for value, n in c.items()]
    c.sort(reverse=True)
    return c

def test():
    size = 512
    for r in (0, 63, 127, 191, 255):
        im = PIL.Image.new("RGB", (size, size))
        for g in range(0, size):
            for b in range(0, size):
                value = get_value((r, (g * 256) / size, (b * 256 / size)))
                im.putpixel((g, b), value)
        im.save("rgb%d.png" % r)

def test_flat(rgb):
    size = 64
    im = PIL.Image.new("RGB", (size, size))
    for y in range(0, size):
        for x in range(0, size):
            im.putpixel((x, y), get_value(rgb))
    im.save("rgb%02d%02d%02d.png" % rgb)

def rotate_and_scale(exif, im, width, height, rotate):
    if rotate or exif and exif["Image Orientation"].values == [6L]:
        im = im.rotate(270)

    w, h = im.size
    if w > h:
        height = (width * h) / w
    else:
        width = (height * w) / h

    return im.resize((width, height))

def count_colours(im, colours):
    width, height = im.size
    for y in range(0, height):
        l = set()
        for x in range(0, width):
            l.add(im.getpixel((x, y)))
        if len(l) > colours:
            return (y, l)
    return None

if __name__ == "__main__":
    if "--test" in sys.argv:
        test()
        sys.exit(0)
    elif "--test-flat" in sys.argv:
        test_flat((120, 40, 60))
        sys.exit(0)

    width = 320
    height = 256

    input_filename, output_filename = sys.argv[1:3]
    basename, ext = splitext(output_filename)
    preview_filename = "".join([basename + "_preview", ext])

    options = sys.argv[3:]

    rotate = "-r" in options
    saturate = options.count("-s")
    desaturate = options.count("-d")
    darken = options.count("-D")
    brighten = options.count("-B")
    square = "-2" in options and square or (lambda x: x)
    preview = "-p" in options
    half_resolution_preview = "-h" in options
    verify = "-v" in options
    no_normal_output = "-n" in options
    make_image = not no_normal_output

    if make_image or preview:
        exif = EXIF.process_file(open(input_filename))
        im = PIL.Image.open(input_filename).convert("RGB")
        im = rotate_and_scale(exif, im, width, height, rotate)

        width, height = im.size

        if saturate or desaturate or darken or brighten:
            for y in range(0, height):
                for x in range(0, width):
                    if saturate or desaturate:
                        rgb = im.getpixel((x, y))
                        rgb = saturate_rgb(rgb, saturate and math.pow(0.5, saturate) or math.pow(2, desaturate))
                        im.putpixel((x, y), rgb)

                    if darken or brighten:
                        rgb = im.getpixel((x, y))
                        rgb = amplify_rgb(rgb, brighten and math.pow(0.5, brighten) or math.pow(2, darken))
                        im.putpixel((x, y), rgb)

    if preview:
        imp = im.copy()
        step = half_resolution_preview and 2 or 1
        for y in range(0, height):
            for x in range(0, width, step):
                rgb = imp.getpixel((x, y))
                value = get_value(rgb)
                imp.putpixel((x, y), value)
                if half_resolution_preview:
                    imp.putpixel((x+1, y), value)

        imp.save(preview_filename)

    if make_image:
        for y in range(0, height):
            c = get_colours(im, y)
            most = [value for n, value in c[:4]]
            least = [value for n, value in c[4:]]

            for x in range(0, width):
                rgb = im.getpixel((x, y))
                value = get_value(rgb, most)
                im.putpixel((x, y), value)

                if y < height - 1:
                    rgbn = im.getpixel((x, y+1))
                    rgbn = tuple(map(lambda i: clip(i[0] + i[1] - i[2]), zip(rgbn, rgb, value)))
                    im.putpixel((x, y+1), rgbn)

        im.save(output_filename)

    if verify:
        if no_normal_output:
            im = PIL.Image.open(output_filename).convert("RGB")

        result = count_colours(im, 4)
        if result is not None:
            y, colours = result
            print "Row %d has the following colours: %s" % (y, "; ".join([repr(c) for c in colours]))

# vim: tabstop=4 expandtab shiftwidth=4
